#include "Aligner.h"
#include "AffixIterator.h"
#include "Sequence.h"
#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <utility>

using namespace std;

namespace opt {
	/** For a duplicate k-mer in the target
	 * ERROR: report an error and exit
	 * MULTIMAP: report all alignments
	 * IGNORE: do not report any alignments
	 */
	extern int verbose; // defined in COMMON.
	int multimap;
	static unsigned k;
	unsigned a;
	SimpleContigGraph contigGraph;
	extern int multiples; // defined in KAligner.
};

      std::ostream& operator <<(std::ostream& out,
                        const std::deque<ExtAlignment >& a)
        {
                bool first = true;
                for (std::deque<ExtAlignment >::const_iterator alignIter = a.begin(); alignIter != a.end(); ++alignIter){
                        if (!first)
                                out << ',';
                        first = false;
                        out << alignIter->second.contig 
                            << (alignIter->second.isRC ? '-':'+');
                }
                return out;
        }


template <>
void Aligner<SeqPosHashMultiMap>::addReferenceSequence(
		const Kmer& kmer, Position pos)
{
	assert(opt::multimap == opt::MULTIMAP);
	m_target.insert(make_pair(kmer, pos));
}

template <class SeqPosHashMap>
void Aligner<SeqPosHashMap>::addReferenceSequence(
		const Kmer& kmer, Position pos)
{
	assert(opt::multimap != opt::MULTIMAP);
	Kmer rc_kmer = reverseComplement(kmer);
	map_iterator findIt = m_target.find(rc_kmer);

	if (findIt != m_target.end()) {
		if (!findIt->second.isDuplicate())
			findIt->second.setDuplicate(
					contigIndexToID(findIt->second.contig),
					contigIndexToID(pos.contig), kmer.decode());
		return;
	}

	pair<map_iterator, bool> inserted
		= m_target.insert(make_pair(kmer, pos));

	if (!inserted.second)
		if (!inserted.first->second.isDuplicate())
			inserted.first->second.setDuplicate(
					contigIndexToID(inserted.first->second.contig),
					contigIndexToID(pos.contig), kmer.decode());
}

/** Create an index of the target sequence. */
template <class SeqPosHashMap>
void Aligner<SeqPosHashMap>::addReferenceSequence(
		const ContigID& idString, const Sequence& seq)
{
	unsigned id = contigIDToIndex(idString);
	
	 //store contig sequences 
	// Sequence* seqstored = new Sequence(seq);
	// const Sequence* seqstored = &seq;
	//const Sequence* seqstored = seq;
	//const Sequence& seqstored = seq;
	//m_contigSeqDict.push_back(seqstored);
	m_contigSeqDict.push_back(new Sequence(seq));

	int size = seq.length();
	for(int i = 0; i < (size - m_hashSize + 1); ++i)
	{
		Sequence subseq(seq, i, m_hashSize);
		if (subseq.find("N") != string::npos)
			continue;
		addReferenceSequence(Kmer(subseq), Position(id, i));
	}
}

/** Create an index of the target sequence. */
//template <class SeqPosHashMap>
//void Aligner<SeqPosHashMap>::addReferenceSequence(
//		const ContigID& idString, const Sequence& seq)
//{
//	unsigned id = contigIDToIndex(idString);
//	int size 
	

template <class SeqPosHashMap>
template <class oiterator>
void Aligner<SeqPosHashMap>::alignRead(const Sequence& seq, unsigned k,
		oiterator dest)
{
	//std::cout << opt::contigGraph.getNumVertices();
	opt::k = k;

	// This vector gathers the (possibly multiple) alignment paths before coalescing them
	std::vector<AlignmentPath> alignBest;
	Sequence rcseq = reverseComplement(seq);
	getAlignmentsInternal(seq, rcseq, false, alignBest, opt::a, opt::contigGraph);
	getAlignmentsInternal(seq, rcseq, true, alignBest, opt::a, opt::contigGraph);

	if (alignBest.empty()) {
		// cout << "found no alignment path, bailing out" << endl;
		return;
	}
	
	// The results to report
	AlignmentSet aligns;
	vector<ExtAlignment> alignPrototype;
	for (vector<AlignmentPath>::iterator iter = alignBest.begin();
		iter != alignBest.end(); iter++)
		if (opt::multiples) 
			// can just add all the alignments in the path
			addAlignments(aligns, *iter);
		else if (!iter->unique()) {
			// cout << "got multiple different alignment paths, bailing out" << endl;
			return;
		}
		// now it gets tricky:
		// If this is the first alignment path, make a vector out of it,
		// so we can compare it with other paths.
		else if (alignPrototype.empty())
			alignPrototype = *iter; // AlignmentPath converted to vector
		// Compare every subsequent alignment path. It may have been derived
		// using a different contig as seed, but the comparison operation is 
		// ignorant of that.
		// But if the contig sequence is different, bail out.
		else if (!(*iter == alignPrototype)) {
			// cout << "got multiple different alignment paths, bailing out" << endl;
			return;
		}
	// end for
	if (!opt::multiples) 
		// we have survived the uniqueness check, but we must actually read out
		// our unique path.
		addAlignments(aligns, alignBest[0]);

	//std::cout << "coalescing ..." << endl;
	coalesceAlignments(aligns, dest);
	//std::cout << "end of colescing ..." << endl;
}

template <class SeqPosHashMap>
void Aligner<SeqPosHashMap>::printContig(unsigned ctgIndex, Sequence *seq, int ctgLen, int start, int len)
{
       	if (opt::verbose >= 4) {
	    if (start < 0) {len += start; start = 0;}
	    if (start + len > ctgLen) len = ctgLen - start;
	    Sequence sseq = seq->substr(start, len);
	    // std::cout << " ID: " << contigIndexToID(ctgIndex) << " (" << start << " to " << start+len << " of " << ctgLen << ")" << endl;
	    std::cout << " Contig Sequence: " << sseq << " ID: " << contigIndexToID(ctgIndex) << " (" << start << " to " << start+len << " of " << ctgLen << ")" << endl;
	}
}

template <class SeqPosHashMap>
unsigned Aligner<SeqPosHashMap>::getMismatches (Sequence::const_iterator first, int compareLength, Sequence::const_iterator second)
{
	Sequence::const_iterator last = first + compareLength;
	if (opt::verbose >= 4)
		std::cout << "read_start: " << *first << " "
		 	<< "read_end: " << *last << " "
		 	<< "ctg_start: " << *second;
	pair<Sequence::const_iterator, Sequence::const_iterator> mypair (first, second);
	unsigned mismatches = 0;
	mypair = mismatch(mypair.first, last, mypair.second);
	while (mypair.first != last)
	{
			mypair.first++;
			mypair.second++;
			mismatches++;
			mypair = mismatch (mypair.first, last, mypair.second);
	}
	return mismatches;
}

template <class SeqPosHashMap>
void Aligner<SeqPosHashMap>::
alignReadLeftDFS(SimpleContigGraph::VertexType* pCurrVertex, AlignmentPath& align, const Sequence& seq, const Sequence& rcseq, unsigned overhang, int overlap, bool isRC)
{
	extDirection dir = isRC ? ANTISENSE : SENSE;

	if (opt::verbose >= 4) 
		std::cout << "LeftDFS with overlap: " << overlap
			<< " overhang: "<< overhang << " extDirection: " << dir
			<< " SENSE: " << SENSE << endl;

	//std::cout << " node.isRC : " << eIter->reverse;
        //std::cout << " : " << seq;
        //std::cout << " : " << complete;
	//std::cout << " : " << isRC;


	
        // Recursively explore the subbranches until either the constraints have been violated or all the constraints have been satisfied.
	typename SimpleContigGraph::VertexType::EdgeCollection& currEdges = pCurrVertex->m_edges[!dir];
	for (typename SimpleContigGraph::VertexType::EdgeCollection::iterator eIter = currEdges.begin(); eIter != currEdges.end(); ++eIter)
	{
		SimpleContigGraph::VertexType* pNextVertex = eIter->pVertex;
		// get the contig index
		unsigned ctgIndx = pNextVertex->m_key;
		bool reverse = eIter->reverse;
	
		unsigned overhang_next = overhang;
		// bool rc = dir ^ reverse;
		bool rc = ((dir == SENSE) == reverse);
		unsigned mismatches;
		if (compare(true, align, rc?rcseq:seq, ctgIndx, overhang_next, overlap, rc, mismatches))
		{
			// int aligned_length = ctg_left_length;
			// Alignment align(contigIndexToID(ctgIndx), ctg_start_coord, read_start_coord, aligned_length, seqLen, isRC);
			// If verbose >= 3, report all (including intermediate) paths; otherwise, only report complete ones
			if (opt::verbose >= 2 && overhang_next == 0 || opt::verbose >= 3)
				std::cout << "AlignmentLeft: " << align.currentExtension << endl;
			if (overhang_next == 0)
				// Indicate that we have found at least one complete match
				align.finishLeft();
			// Not complete yet? Then recurse!
			else 
				alignReadLeftDFS(pNextVertex, align, seq, rcseq, overhang_next, overlap, isRC);
			// align.align_length -= compare_length;
			// align.read_start_pos += compare_length;
			// restore old alignment and # of mismatches
			align.popLeft(mismatches);
		}
	}
}

// This compares up to overhang_next positions of the ORIGINAL sequence:
// - from the LEFT/RIGHT, if leftright==true/false
// - with the given contig/its reverse complement, if rc == false/true 
template <class SeqPosHashMap>
bool Aligner<SeqPosHashMap>::
compare(bool leftright, AlignmentPath& align, const Sequence& qseq, unsigned ctgIndx, unsigned& overhang_next, int overlap, bool rc, unsigned& mismatches)
{
	Sequence* ctgseq = contigIndexToSeq(ctgIndx);
	unsigned ctgLen = ctgseq->length();
	unsigned compare_length = ctgLen - overlap;
	if (compare_length > overhang_next)
		compare_length = overhang_next;
	unsigned ctg_start_coord, read_start_coord;
	if (leftright == rc) {
		ctg_start_coord = overlap;
		read_start_coord = qseq.length() - overhang_next;
	} else {
		ctg_start_coord = ctgLen - overlap - compare_length;
		read_start_coord = overhang_next - compare_length;
	}
	if (opt::verbose >= 4)
	       	std::cout << " ctg_start_coord: " << ctg_start_coord 
		<< " read_start_coord: " << read_start_coord 
               	<< " compare_length: " << compare_length << endl
		<< "read seq to: " << qseq.substr(read_start_coord, compare_length) << std::endl
		<< "compare with:" << ctgseq->substr(ctg_start_coord, compare_length) << std::endl;

	overhang_next -= compare_length;
	// only compare the non-overlapping portion
	mismatches = getMismatches(
		qseq.begin() + read_start_coord,
		compare_length,
		ctgseq->begin() + ctg_start_coord);
	if (opt::verbose >= 4)
		std::cout << " overhang_next: " << overhang_next 
		<< " mismatches: " << mismatches << endl;
		
	// report the alignment, and update the mismatches, if max number not exceeded.
	return leftright 
	? align.pushLeft(ctgIndx, contigIndexToID(ctgIndx), rc ? 0 : ctgLen - compare_length - overlap, overhang_next, compare_length + overlap, rc, mismatches)
	: align.pushRight(ctgIndx, contigIndexToID(ctgIndx), rc ? ctgLen - compare_length - overlap : 0, qseq.length() - overhang_next - compare_length - overlap, compare_length + overlap, rc, mismatches);
}

template <class SeqPosHashMap>
void Aligner<SeqPosHashMap>::
alignReadRightDFS(SimpleContigGraph::VertexType* pCurrVertex, AlignmentPath& align, const Sequence& seq, const Sequence& rcseq, unsigned overhang, int overlap, bool isRC)
{
	extDirection dir = isRC ? ANTISENSE : SENSE;

	if (opt::verbose >= 4) {
		std::cout << "RightDFS with overlap: " << overlap
			<< " overhang: "<< overhang << " extDirection: " << dir
			<< " SENSE: " << SENSE << endl;
	}
	// Recursively explore the subbranches until either the constraints have been violated or all the constraints have been satisfied.
	typename SimpleContigGraph::VertexType::EdgeCollection& currEdges = pCurrVertex->m_edges[dir];
	for(typename SimpleContigGraph::VertexType::EdgeCollection::iterator eIter = currEdges.begin(); eIter != currEdges.end(); ++eIter)
	{
		SimpleContigGraph::VertexType* pNextVertex = eIter->pVertex;
		
		// get the contig index
		unsigned ctgIndx = pNextVertex->m_key;
		bool reverse = eIter->reverse;
		if (opt::verbose >= 4) std::cout << " reverse? " << reverse;

		bool rc = dir ^ reverse;	
		unsigned overhang_next = overhang;
		unsigned mismatches;
		if (compare(false, align, rc?rcseq:seq, ctgIndx, overhang_next, overlap, rc, mismatches))
		{
			//int aligned_length = overlap + ctg_right_length;
			// int aligned_length = ctg_right_length;
			// add the alignments to the temporary result AlignmentSet
		 	//std::cout << " chosen Contig ID: " << contigIndexToID(ctgIndx);
			// If verbose >= 3, report all (including intermediate) paths; otherwise, only report complete ones
			if (opt::verbose >= 2 && overhang_next == 0 || opt::verbose >= 3)
				std::cout << "AlignmentRite: " << align.currentExtension << endl;
			if (overhang_next == 0)
				// Indicate that we have found at least one complete match
				align.finishRight();
			//std::cout << "overhang_next: " << overhang_next << endl;
			// Not complete yet? Then recurse!
			else 
				alignReadRightDFS(pNextVertex, align, seq, rcseq, overhang_next, overlap, isRC);
			// align.align_length -= compare_length;
			// restore old align and # of mismatches
			align.popRight(mismatches);
		}				
	}
}
			
template <class SeqPosHashMap>
void Aligner<SeqPosHashMap>::
getAlignmentsInternal(const Sequence& seq, const Sequence& rcseq, bool isRC,
		std::vector<AlignmentPath>& alignBest, unsigned mismatchAllowed, const SimpleContigGraph& graph)
{

	const Sequence &qseq = isRC?rcseq:seq;
	if (opt::verbose >= 4) std::cout << "  Query sequence: " << qseq << endl;
	bool good = qseq.find_first_not_of("ACGT0123") == string::npos;
	bool discarded = true;
	int seqLen = seq.length();
	
	// Seed number is decided by number of mismatches allowed
	int seed_num = mismatchAllowed + 1;

	// Seed length is decided by the number of mismatches allowed 
	// perform soft clip
	
	int seed_length = seqLen/(mismatchAllowed+1);

	//std::cout << "seed_length: " << seed_length << endl;

	//for(int i = 0; i < (seqLen - m_hashSize) + 1; i++)
	//int i = 0;
	//bool found = false;

	// mm gives us the minimum number of mismatches found so far.
	// Initalized to mismatchAllowed+1, i.e. no alignments found so far.
	// Once we have found at least one alignment, (i.e. mm has become
	// <= mismatchAllowed), we set mismatchAllowed down to mm,
	// i.e for the rest of the search we are only interested in aligments
	// with at most mm mismatches. Same each time we beat our previous minimum
	// number of mismatches.
	// unsigned mm = mismatchAllowed + 1;
	for (int i = 0; i < seed_num; i++)
	//while ((i < seed_num) and (!found))
	// while (i < seed_num)
	{
		//pair<map_const_iterator, map_const_iterator> range
		//	= m_target.equal_range(Kmer(qseq.substr(i, m_hashSize)));
		int seed_start = i * seed_length;
		if (opt::verbose >= 4) std::cout << "> Seed: " << seed_start << endl;
		Sequence kmer(qseq.substr(seed_start, seed_length));
		if (!good && kmer.find_first_not_of("ACGT0123") != string::npos)
			continue;
		discarded = false;

		pair<map_const_iterator, map_const_iterator> range
			= m_target.equal_range(Kmer(qseq.substr(seed_start, seed_length)));
		//int max_i = i;
		for (map_const_iterator resultIter = range.first;
				resultIter != range.second; ++resultIter) {
			if (opt::multimap == opt::IGNORE
					&& resultIter->second.isDuplicate())
				break;

			//int read_pos = !isRC ? seed_start
			//	: Alignment::calculateReverseReadStart(
			//			seed_start, seqLen, seed_length);
			unsigned ctgIndex = resultIter->second.contig;
			if (opt::verbose >= 4) 
			//std::cout << " seed_start: " << seed_start << endl;
			//std::cout << " read_pos: " << read_pos << " ";
			//std::cout << " isRC: " << isRC << " ";
				std::cout << ">> ctgIndex: " << contigIndexToID(ctgIndex) << " ";
			//Alignment align(contigIndexToID(ctgIndex),
			//		resultIter->second.pos, read_pos, m_hashSize,
			//		seqLen, isRC);
			// std::cout << " Hash match: " << align;
			//aligns[ctgIndex].push_back(align);
			Sequence* ctgseq = contigIndexToSeq(ctgIndex);
			//std::cout << "Level 0 contig seq: "<< *ctgseq << endl;
			int ctgLen = ctgseq->length();
			//std::cout << " Contig Length: " << ctgLen;
			int ctg_pos = resultIter->second.pos;

			SimpleContigGraph::VertexType* pSourceVertex;
			pSourceVertex = graph.findVertex(ctgIndex);
			printContig(ctgIndex, ctgseq, ctgLen, ctg_pos - seed_start, seqLen);

			// if (opt::verbose >= 4) std::cout << "Contig ID: " << contigIndexToID(pSourceVertex->m_key) << endl;
			
			//std::cout << "ctg_pos: " << ctg_pos << " ";
			//extend
			//int j = i + 1;
			//int j = i;
			//keep track of number of mismatches
			unsigned m = 0;
			//while (j < (seqLen - m_hashSize) + 1)
			int seed_right_start = seed_start + seed_length; // read_pos
			int seed_right_end = seqLen;
			int seed_right_length = seed_right_end - seed_right_start;

			int seed_left_start = 0;
			//int seed_left_end = read_pos;
			//int seed_left_length = seed_left_end - seed_left_start;

			int ctg_right_start = ctg_pos + seed_length; 
			int ctg_right_end = ctg_right_start + seed_right_length;
			if (opt::verbose >= 4)
				std::cout << "ctg_right_start: " << ctg_right_start 
				<< " ctg_right_end: " << ctg_right_end 
				<< " ctgLen: " << ctgLen << std::endl;

			int overlap = opt::k - 1;
			// the overhang variables are wrt. the ORIGINAL seq, NOT wrt. qseq.
			unsigned overhang_right = 0;
			unsigned overhang_left = 0;
			int ctg_right_length;
			if (ctg_right_end > ctgLen)
			{
				//m = m + (ctg_right_end - ctgLen);

				ctg_right_length = ctgLen - ctg_right_start;
				
				if (isRC)
					overhang_left = ctg_right_end - ctgLen;
				else
					overhang_right = ctg_right_end - ctgLen;
				ctg_right_end = ctgLen;
				//std::cout << "overhang_right: " << overhang_right << endl;
				//int unaligned_length = ctg_right_end - ctgLen;
				//std::cout << "unaligned_length " << unaligned_length;
				//alignReadRightDFS(pSourceVertex, aligns, seq, rcseq, unaligned_length, mismatchAllowed, m, overlap, false);
				//m = m + (ctg_right_end - ctgLen);
			}
			else
				ctg_right_length = ctg_right_end - ctg_right_start;
			//std::cout << "overhang_right: " << overhang_right << endl;
			//std::cout << "seed_start: " << seed_start << " ";
			//std::cout << "read_pos: " << read_pos << " ";
			//int ctg_left_start = ctg_pos - seed_start;
			int ctg_left_start = ctg_pos - seed_start; // read_pos
			int ctg_left_end = ctg_pos;
			if (opt::verbose >= 4)
				std::cout << "ctg_left_start: " << ctg_left_start 
				<< " ctg_left_end: " << ctg_left_end << std::endl;
			int ctg_left_length;
			if (ctg_left_start < 0)
			{
				//m = m + (0 - ctg_left_start);
				ctg_left_length = ctg_left_end - 0;
				if (isRC)
					overhang_right = 0 - ctg_left_start;
				else
					overhang_left = 0 - ctg_left_start;
				ctg_left_start = 0;
				seed_left_start = seed_start - ctg_left_length;
			} else 
				ctg_left_length = ctg_left_end - ctg_left_start;

			//std::cout << "mismatchAllowed " << mismatchAllowed << " ";
			//std::cout << "m " << m << endl;
			if (m > mismatchAllowed)
				continue;
			//std::cout << "seed_right_start: " << seed_right_start << " ";
			//std::cout << "ctg_right_length: " << ctg_right_length << " ";
			//std::cout << "qseq.begin(): " << qseq.begin() << " ";

			// right-extension
			m += getMismatches(
				qseq.begin() + seed_right_start,
				ctg_right_length,
				ctgseq->begin() + ctg_right_start);
			if (m > mismatchAllowed)
				continue;

			//std::cout << "seed_left_start: " << seed_left_start << " ";
			//std::cout << "ctg_left_length: " << ctg_left_length << " ";
			//std::cout << qseq.substr(seed_left_start, ctg_left_length) << ':'
			//<< ctgseq->substr(ctg_left_start, ctg_left_length) << ';';
			// left-extension
			m += getMismatches(
				qseq.begin() + seed_left_start,
				ctg_left_length,
				ctgseq->begin() + ctg_left_start);
			if (m > mismatchAllowed)
				continue;

			int aligned_length = ctg_right_end - ctg_left_start;
			// if no more than overlap basepairs match, we can skip it, because
			// there must be an adjacent contig that 
			// completely subsumes this match, i.e. it matches more positions
			if (aligned_length <= overlap)
				continue;

			// set up a temporary result AlignmentPath, with the seed ctg in it
			AlignmentPath align(ctgIndex, contigIndexToID(ctgIndex), ctg_left_start, overhang_left, aligned_length, seqLen, isRC, m, mismatchAllowed);
			// all the best paths found so far
			// alignBest.push_back(align);
			// AlignmentSet alignsTmp;
			// Alignment align(contigIndexToID(ctgIndex), ctg_left_start, seed_left_start, aligned_length, seqLen, isRC);
			// alignsTmp[ctgIndex].push_back(align);
					
			//if (overhang_left > 0)
			//{
			//	alignReadLeftDFS(pSourceVertex, alignsTmp, align, seq, rcseq, overhang_left, mismatchAllowed, m, overlap, false, isRC);
			//}
				
			//for (AlignmentSet::const_iterator ctgIter = alignsTmp.begin(); ctgIter != alignsTmp.end(); ++ctgIter) {
			//	AlignmentVector alignVec = ctgIter->second;
			//	std::cout << endl << "Contig ID: " << contigIndexToID(ctgIter->first);
			//	for (AlignmentVector::iterator alignIter = alignVec.begin(); alignIter != alignVec.end(); ++alignIter) {
			//		Alignment currAlign = *alignIter;
			//		std::cout << "Alignment: " << "Contig start: " << currAlign.contig_start_pos << endl;
			//		std::cout << "read_start_pos: " << currAlign.read_start_pos << endl;
			//		}
			//	}
			//std::cout << "aligned_length: " << currAlign.aligned_length << endl; 
			if (opt::verbose >= 4) {
				std::cout << "overhang_right: " << overhang_right << endl;
				std::cout << "overhang_left: " << overhang_left << endl;
			}
			if (overhang_right > 0)
			{
				align.startRight();
				alignReadRightDFS(pSourceVertex, align, seq, rcseq, overhang_right, overlap, isRC);	
				if (opt::verbose >= 4) std::cout << "complete_right: " << align.nonEmpty() << endl;
			}
			if (overhang_left > 0 && align.nonEmpty())
			{
				align.startLeft();
				alignReadLeftDFS(pSourceVertex, align, seq, rcseq, overhang_left, overlap, isRC);	
				if (opt::verbose >= 4) std::cout << "complete_left: " << align.nonEmpty() << endl;
			}
				
			if (align.nonEmpty())
			{
				// Verbose output: then print the alignment path(s) found
				if (opt::verbose >= 2) 
					std::cout << "Best Alignment: " << endl << align << endl;
				// Note: For efficiency, we could refuse to add this alignment path,
				// and abandon alignment with this number of mismatches
				// if the path is not unique. But we would need to "earmark" 
				// this somehow.
				unsigned mm_new = align.getMismatches();
				if (mm_new < mismatchAllowed) {
					alignBest.clear();
					mismatchAllowed = mm_new;
				}
				// record this alignment path
				alignBest.push_back(align);
			}
		}
			 
	}

}


//static int compareQueryPos(const Alignment& a1, const Alignment& a2)
//{
//	return a1.read_start_pos < a2.read_start_pos;
//}

static int compareTargetPos(const Alignment& a1, const Alignment& a2)
{
	return a1.contig_start_pos < a2.contig_start_pos;
}

/** Coalesce the k-mer alignments into a read alignment. */
template <class SeqPosHashMap>
template <class oiterator>
void Aligner<SeqPosHashMap>::
coalesceAlignments(const AlignmentSet& alignSet, oiterator& dest)
{
	for (AlignmentSet::const_iterator ctgIter = alignSet.begin();
			ctgIter != alignSet.end(); ++ctgIter) {
		AlignmentVector alignVec = ctgIter->second;
		//unsigned ctgIndex = ctgIter->first;
		//std::cout << "ctgIndex: " << ctgIndex << endl;
		//std::cout << "alignVec: " << alignVec.begin()->read_start_pos << endl;
		assert(!alignVec.empty());

		//sort(alignVec.begin(), alignVec.end(), compareQueryPos);
		sort(alignVec.begin(), alignVec.end(), compareTargetPos);

		AlignmentVector::iterator prevIter = alignVec.begin();
		AlignmentVector::iterator currIter = alignVec.begin() + 1;
		Alignment currAlign = *prevIter;
		//AlignmentVector::iterator currIter = alignVec.begin();
		//AlignmentVector::iterator result = currIter;
		while (currIter != alignVec.end()) {
			//int qstep = 1;
			//int tstep = currIter->isRC ? -1 : 1;
			//if (currIter->read_start_pos
			//			!= prevIter->read_start_pos + qstep
			//		|| currIter->contig_start_pos
			//			!= prevIter->contig_start_pos + tstep) {
			//	*dest++ = currAlign;
			//	currAlign = *currIter;
			//} else {
			//	currAlign.align_length++;
			//	if (currAlign.isRC)
			//		currAlign.contig_start_pos--;
			//}
			//std::cout << prevIter->contig_start_pos << ":" << endl;
			//std::cout << currIter->contig_start_pos << ":" << endl;	
			if (currIter->contig_start_pos != prevIter->contig_start_pos
				|| currIter->read_start_pos != prevIter->read_start_pos
				|| currIter->align_length != prevIter->align_length
				|| currIter->isRC != prevIter->isRC)
			{
				//result++;
				//*result = *currIter;
				//std::cout <<"I'm here" << endl;
				*dest++ = currAlign;
				currAlign = *currIter;
			}
			prevIter = currIter;
			currIter++;
		}
		//std::cout << "I'm here" << endl;
		*dest++ = currAlign;
	}
}

// Explicit instantiation.
template void Aligner<SeqPosHashMultiMap>::addReferenceSequence(
		const ContigID& id, const Sequence& seq);

template void Aligner<SeqPosHashUniqueMap>::addReferenceSequence(
		const ContigID& id, const Sequence& seq);

template void Aligner<SeqPosHashMultiMap>::
alignRead<affix_ostream_iterator<Alignment> >(
		const Sequence& seq, unsigned k, affix_ostream_iterator<Alignment> dest);

template void Aligner<SeqPosHashUniqueMap>::
alignRead<affix_ostream_iterator<Alignment> >(
		const Sequence& seq, unsigned k, affix_ostream_iterator<Alignment> dest);
