#ifndef ALIGNER_H
#define ALIGNER_H 1

//
// Aligner - Simple class to do an approximate alignment of reads to the input reference sequence
//

#include "Options.h"
#include "config.h"
#include "HashMap.h"
#include "Kmer.h"
#include "ContigGraph.h"
// #include "DirectedGraphImpl.h"
#include <cassert>
#include <cstdlib>
#include <iostream>
#include <istream>
#include <limits>
#include <map>
#include <ostream>
#include <string>
#include <deque>
#include <vector>

typedef std::string ContigID;


// Alignment information
struct Alignment
{
	ContigID contig;
	int contig_start_pos;
	int read_start_pos;
	int align_length;
	int read_length;
	bool isRC;

	Alignment() { }

	Alignment(ContigID contig, int contig_start, int read_start,
			int align_length, int read_length, bool isRC) :
		contig(contig),
		contig_start_pos(contig_start),
		read_start_pos(read_start),
		align_length(align_length),
		read_length(read_length),
		isRC(isRC)
	{
	}

	/**
	 * Return the taret position at the query start.
	 * Note: not alignment start, and may be negative
	 */
	int targetAtQueryStart() const
	{
		unsigned tend = contig_start_pos + align_length;
		return !isRC ? contig_start_pos - read_start_pos
			: tend + read_start_pos;
	}

	/** Return the distance between the specified alignments.
	 * May be used to calculate fragment size when the alignments are
	 * mate pairs.
	 */
	int operator-(const Alignment& o) const
	{
		return targetAtQueryStart() - o.targetAtQueryStart();
	}

	// Comparison of two contigs
	bool operator==(const Alignment& o) const
	{
		return contig==o.contig 
			&& contig_start_pos==o.contig_start_pos
                	&& read_start_pos==o.read_start_pos
                	&& align_length==o.align_length
                	&& read_length==o.read_length
                	&& isRC==o.isRC;
	}

	bool operator!=(const Alignment& o) const { return !operator==(o); }

	/** This alignment is converted to the corresponding alignment of
	 * the same query to the reverse complement of the target.
	 */
	Alignment flipTarget(unsigned tlength) const
	{
		Alignment rc(*this);
		unsigned tend = contig_start_pos + align_length;
		assert(tend <= tlength);
		rc.contig_start_pos = tlength - tend;
		rc.isRC = !isRC;
		return rc;
	}

	/** Return an alignment of the reverse complement of the query to
	 * the same target.
	 */
	Alignment flipQuery() const
	{
		Alignment rc(*this);
		unsigned qend = read_start_pos + align_length;
		assert(qend <= (unsigned)read_length);
		rc.read_start_pos = read_length - qend;
		rc.isRC = !isRC;
		return rc;
	}

	static int calculateReverseReadStart(int read_start_pos,
			int read_length, int align_length)
	{
		unsigned qend = read_start_pos + align_length;
		return read_length - qend;
	}

	friend std::istream& operator >>(std::istream& in, Alignment& a)
	{
		return in >> a.contig
			>> a.contig_start_pos
			>> a.read_start_pos
			>> a.align_length
			>> a.read_length
			>> a.isRC;
	}

	friend std::ostream& operator <<(std::ostream& out,
			const Alignment& a)
	{
		return out << a.contig << ' '
			<< a.contig_start_pos << ' '
			<< a.read_start_pos << ' '
			<< a.align_length << ' '
			<< a.read_length << ' '
			<< a.isRC;
	}
};

typedef std::vector<Alignment> AlignmentVector;
// Sadly, there is no way to retrieve contig indices from contigs,
// and the Alignment data structure does not store contig indices,
// so we need to store them ourselves
typedef std::pair<unsigned, Alignment> ExtAlignment;

// a multi-alignment along a path of contigs
struct AlignmentPath {
	ExtAlignment seedContig;
	std::deque<ExtAlignment > currentExtension;

	std::vector<std::deque<ExtAlignment > > leftExtensions;
	std::vector<std::deque<ExtAlignment > > rightExtensions;
	
	// this indicates that either the seed has not yet been extended in the particular
	// direction, or it need not be extended in that direction, or it has been successfully
	// extended (with at least one extension.
	bool leftComplete, rightComplete;

	// Mismatch counters
	unsigned mmAllowed;
	// mismatches encountered in the seed contig
	unsigned seedMM;
	// mismatches encountered in the current extension
	unsigned currentExtensionMM;
	// minimum number of MM in the left and right extensions so far
	unsigned minLeftMM, minRightMM;

	AlignmentPath(unsigned seedIndex, ContigID contig, int contig_start_pos, int read_start_pos, int align_length, int read_length, bool isRC, unsigned seedmm, unsigned mm_allow) :
		seedContig(seedIndex, Alignment(contig, contig_start_pos, read_start_pos, align_length, read_length, isRC)),
		leftComplete(true),
		rightComplete(true),
		mmAllowed(mm_allow),
		seedMM(seedmm),
		currentExtensionMM(0),
		// these values are correct if no left/right extension is necessary, resp.
		minLeftMM(0),
		minRightMM (0) {}

	// This is called at most once in the lifetime of an AlignmentPath, namely
	// right after initialization, if the seed contig needs to be extended to the right.
	void startRight(void) 
	{
		// can have at most this many mm on the right
		minRightMM = mmAllowed - seedMM;
		rightComplete = false;
	}
	// The push methods return false (and don't push anything) if the given 
	// number of mismatches makes the current extension exceed the extension's minimum
	// number of mismatches (i.e. the solution is poorer than an existing extension).
	// Otherwise they form an alignment from the given data and append it to the given
	// extension.
	bool pushRight (unsigned contigIndex, ContigID contig, int contig_start_pos, int read_start_pos, int align_length, bool isRC, unsigned mm)
	{
		if (currentExtensionMM + mm > minRightMM)
			return false;
		currentExtensionMM += mm;
		Alignment entry(contig, contig_start_pos, read_start_pos, align_length, seedContig.second.read_length, isRC);
		currentExtension.push_back(ExtAlignment(contigIndex, entry));
		return true;
	}

	// simply an undo for the preceding push operation --- for backtracking
	void popRight(unsigned mm) 
	{
		currentExtensionMM -= mm;
		currentExtension.pop_back();
	}

	void finishRight(void) 
	{
		if (currentExtensionMM < minRightMM)
			// beats all that have gone before
			rightExtensions.clear();
		// push yourself on
		rightExtensions.push_back(currentExtension);
		rightComplete = true;
	}

	// This is called at most once in the lifetime of an AlignmentPath, namely
	// after the RHS DFS is finished and the seed needs to be extended to the left.
	void startLeft(void) 
	{
		// our RHS backtracking must have finished cleanly
		assert(currentExtension.empty() && currentExtensionMM == 0);
		// can have at most this many mm on the left
		minLeftMM = mmAllowed - seedMM - minRightMM;
		leftComplete = false;
	}

	// The push methods return false (and don't push anything) if the given 
	// number of mismatches makes the current extension exceed the extension's minimum
	// number of mismatches (i.e. the solution is poorer than an existing extension).
	// Otherwise they form an alignment from the given data and append it to the given
	// extension.
	bool pushLeft (unsigned contigIndex, ContigID contig, int contig_start_pos, int read_start_pos, int align_length, bool isRC, unsigned mm)
	{
		if (currentExtensionMM + mm > minLeftMM)
			return false;
		currentExtensionMM += mm;
		Alignment entry(contig, contig_start_pos, read_start_pos, align_length, seedContig.second.read_length, isRC);
		currentExtension.push_front(ExtAlignment(contigIndex, entry));
		return true;
	}

	// simply an undo for the preceding push operation --- for backtracking
	void popLeft(unsigned mm) 
	{
		currentExtensionMM -= mm;
		currentExtension.pop_front();
	}

	void finishLeft(void) 
	{
		if (currentExtensionMM < minLeftMM)
			// beats all that have gone before
			leftExtensions.clear();
		// push yourself on
		leftExtensions.push_back(currentExtension);
		leftComplete = true;
	}

	// We have found at least one alignment using this seed, if on both ends:
	// 	either the seed contig extended all the way to the end
	// 	or there is at least one (complete) extension on that end.
	// NOTE: This can also be queried after right-extending and before left-extending;
	// in this case it indicates that we have successfully right-extended (or did not need to).
	bool nonEmpty(void) { return leftComplete && rightComplete; }

	// A nonempty alignment path is unique, if on both ends, there is at most one
	// extension. (There is none iff the seed contig extended all the way to that end)
	// (If the seed contig extends all the way on both ends, the alignment path is
	// necessarily unique.)
	bool unique(void) 
	{
		return ( leftExtensions.size() <= 1 && rightExtensions.size() <= 1);
	}

	unsigned getMismatches() { return minLeftMM + minRightMM + seedMM; }

	// An AlignmentPath can be copied into a vector of successive aligning contigs
	// from left to right. This chooses the first available left and right extension,
	// respectively, hence is only deterministic if the alignment path is unique.
	operator std::vector<ExtAlignment >() 
	{
		std::vector<ExtAlignment >  vec;
		if ( leftExtensions.size() )
			vec.insert(vec.end(), leftExtensions[0].begin(), leftExtensions[0].end());
		vec.push_back(seedContig);
		if ( rightExtensions.size() )
			vec.insert(vec.end(), rightExtensions[0].begin(), rightExtensions[0].end());
		return vec;
	}

	// The "inverse" to copying an alignment path into a vector. It compares 
	// the first available left and right extension, respectively, (and the seed)
	// with the alignments in the given vector and returns true if all alignments match.
	// Hence this is only deterministic if the alignment path is unique.

	bool operator== (std::vector<ExtAlignment > vec)
	{
		std::vector<ExtAlignment >::iterator iter = vec.begin(); 
		if ( leftExtensions.size() )
		    for (std::deque<ExtAlignment >::iterator leftIter
			= leftExtensions[0].begin(); leftIter != leftExtensions[0].end();
			leftIter++)
			if (iter == vec.end() || *leftIter != *iter++)
				return false;
		if (iter == vec.end() || seedContig != *iter++)
			return false;
		if ( rightExtensions.size() )
		    for (std::deque<ExtAlignment >::iterator rightIter
			= rightExtensions[0].begin(); rightIter != rightExtensions[0].end();
			rightIter++)
			if (iter == vec.end() || *rightIter != *iter++)
				return false;
		// now just compare that we have exhausted the alignment vector
		return (iter == vec.end());
	}

	// The "inverse" to copying an alignment path into a vector. It compares 
	// the first available left and right extension, respectively, (and the seed)

/*	friend std::ostream& operator <<(std::ostream& out,
			const std::deque<ExtAlignment >& a)
        {
                bool first = true;
                for (std::deque<ExtAlignment >::const_iterator alignIter = a.begin(); alignIter != a.end(); ++alignIter){
                        if (!first)
                                out << ',';
                        first = false;
                        out << alignIter->second.contig 
			    << (alignIter->second.isRC ? '-':'+');
		}
		return out;
	}
*/
	friend std::ostream& operator <<(std::ostream& out,
			const AlignmentPath& a)
        {
                for (std::vector<std::deque<ExtAlignment > >::const_iterator alignIter = a.leftExtensions.begin(); alignIter != a.leftExtensions.end(); ++alignIter) {
                                //out << "< " << *alignIter << std::endl; //this does not work with GCC 4.1.2
			out << "< ";
	                bool first = true;
	                for (std::deque<ExtAlignment >::const_iterator alignIter1 = alignIter->begin(); alignIter1 != alignIter->end(); ++alignIter1){
        	                if (!first)
                	                out << ',';
                        	first = false;
	                        out << alignIter1->second.contig 
				    << (alignIter1->second.isRC ? '-':'+');
			}
			out << std::endl;
		}
                out << "= " << a.seedContig.second << std::endl;
                for (std::vector<std::deque<ExtAlignment > >::const_iterator alignIter = a.rightExtensions.begin(); alignIter != a.rightExtensions.end(); ++alignIter) {
                                //out << "> " << *alignIter << std::endl; //this does not work with GCC 4.1.2
			out << "> ";
	                bool first = true;
	                for (std::deque<ExtAlignment >::const_iterator alignIter1 = alignIter->begin(); alignIter1 != alignIter->end(); ++alignIter1){
        	                if (!first)
                	                out << ',';
                        	first = false;
	                        out << alignIter1->second.contig 
				    << (alignIter1->second.isRC ? '-':'+');
			}
			out << std::endl;
		}
                return out << "MM" << a.minLeftMM << ':' << a.seedMM << ':' << a.minRightMM << std::endl;

        }

};

struct Position
{
	uint32_t contig;
	uint32_t pos; // 0 indexed
	Position(uint32_t contig = std::numeric_limits<uint32_t>::max(),
			uint32_t pos = std::numeric_limits<uint32_t>::max())
		: contig(contig), pos(pos) { }
	void setDuplicate(const ContigID& thisContig, const ContigID& otherContig,
			const Sequence& kmer)
	{
		if (opt::multimap == opt::IGNORE)
			contig = std::numeric_limits<uint32_t>::max ();
		else {
			std::cerr << "error: duplicate k-mer in "
				<< thisContig
				<< " also in "
				<< otherContig
				<< ": " << kmer << '\n';
			exit(EXIT_FAILURE);
		}
	}
	bool isDuplicate() const {return contig == std::numeric_limits<uint32_t>::max();}
};

typedef hash_multimap<Kmer, Position, hashKmer> SeqPosHashMultiMap;

#if HAVE_GOOGLE_SPARSE_HASH_SET
# include <google/sparse_hash_map>
typedef google::sparse_hash_map<Kmer, Position,
		hashKmer> SeqPosHashUniqueMap;
#else
typedef hash_map<Kmer, Position, hashKmer> SeqPosHashUniqueMap;
#endif


template <class SeqPosHashMap>
class Aligner
{
	public:
		typedef typename SeqPosHashMap::iterator map_iterator;
		typedef typename SeqPosHashMap::const_iterator
			map_const_iterator;

		Aligner(int hashSize, size_t buckets)
			: m_hashSize(hashSize), m_target(buckets) { }

		Aligner(int hashSize, size_t buckets, float factor)
			: m_hashSize(hashSize)
		{
			m_target.max_load_factor(factor);
			m_target.rehash(buckets);
		}

		void addReferenceSequence(const ContigID& id,
				const Sequence& seq);
		void addReferenceSequence(const Kmer& kmer, Position pos);

		template <class oiterator>
		void alignRead(const Sequence& seq, unsigned k, oiterator dest);

		size_t size() const { return m_target.size(); }
		size_t bucket_count() const
		{
			return m_target.bucket_count();
		}

	private:

		// Internal alignment function, perform the actual alignment
		void getAlignmentsInternal(const Sequence& seq, const Sequence& rcseq, bool isRC, std::vector<AlignmentPath>& alignBest, 
			unsigned mismatchAllowed, const SimpleContigGraph& graph);
		
		typedef std::map<unsigned, AlignmentVector> AlignmentSet;

		template <class oiterator>
		static void coalesceAlignments(const AlignmentSet& alignSet,
				oiterator& dest);

		void printContig(unsigned ctgIndex, Sequence *seq, int ctgLen, int start, int len);
		//void alignReadRightDFS(SimpleContigGraph::VertexType* pCurrVertex, extDirection dir, std::vector<AlignmentPath>& alignsTmp, AlignmentPath& align, const Sequence& seq, int unaligned_length, unsigned mismatchAllowed, unsigned m, int kmer, bool complete);

		bool compare(bool leftright, AlignmentPath& align, const Sequence& qseq, unsigned ctgIndx, unsigned& overhang, int overlap, bool rc, unsigned& mismatches);
		void alignReadLeftDFS(SimpleContigGraph::VertexType* pCurrVertex, AlignmentPath& align, const Sequence& seq, const Sequence& rcseq, unsigned overhang, int kmer, bool isRC);
		void alignReadRightDFS(SimpleContigGraph::VertexType* pCurrVertex, AlignmentPath& align, const Sequence& seq, const Sequence& rcseq, unsigned overhang, int kmer, bool isRC);
		unsigned getMismatches (Sequence::const_iterator first, int compareLength, Sequence::const_iterator second);

		// The number of bases to hash on
		int m_hashSize;

		/** A map of k-mer to contig coordinates. */
		SeqPosHashMap m_target;

		/** A dictionary of contig sequences. */
                std::vector<Sequence*> m_contigSeqDict;
                // std::vector<const Sequence*> m_contigSeqDict;
		
		//const Sequence* contigIndexToSeq(unsigned index)
		Sequence* contigIndexToSeq(unsigned index)
                {
                        return m_contigSeqDict.at(index);
		}

		/** A dictionary of contig IDs. */
		std::vector<ContigID> m_contigDict;

		unsigned contigIDToIndex(const ContigID& id)
		{
			m_contigDict.push_back(id);
			return m_contigDict.size() - 1;
		}

		const ContigID& contigIndexToID(unsigned index)
		{
			return m_contigDict.at(index);
		}
	static void addAlignments (AlignmentSet& aligns,
			const std::deque<ExtAlignment >& a)
        {
                for (std::deque<ExtAlignment >::const_iterator alignIter = a.begin(); alignIter != a.end(); ++alignIter){
			ExtAlignment currAlign = *alignIter;	
                        aligns[currAlign.first].push_back(currAlign.second);	
		}
	}

	static void addAlignments (AlignmentSet& aligns,
			const AlignmentPath& a)
        {
                for (std::vector<std::deque<ExtAlignment > >::const_iterator alignIter = a.leftExtensions.begin(); alignIter != a.leftExtensions.end(); ++alignIter)
			addAlignments(aligns, *alignIter);
		aligns[a.seedContig.first].push_back(a.seedContig.second);
                for (std::vector<std::deque<ExtAlignment > >::const_iterator alignIter = a.rightExtensions.begin(); alignIter != a.rightExtensions.end(); ++alignIter)
			addAlignments(aligns, *alignIter);
        }
};

#endif
